/* -*-c-*-
 *
 * Generic mapping support
 *
 * (c) 2019 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Pyke: the Python Kit for Extensions.
 *
 * Pyke is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Pyke is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Pyke.  If not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "pyke.h"

/*----- Iteration ---------------------------------------------------------*/

static PyTypeObject *itemiter_pytype, *valiter_pytype;

typedef struct iter_pyobj {
  PyObject_HEAD
  PyObject *map;
  PyObject *i;
} iter_pyobj;
#define ITER_MAP(o) (((iter_pyobj *)(o))->map)
#define ITER_I(o) (((iter_pyobj *)(o))->i)

static void iter_pydealloc(PyObject *me)
  { Py_DECREF(ITER_MAP(me)); Py_DECREF(ITER_I(me)); FREEOBJ(me); }

static PyObject *itemiter_pynext(PyObject *me)
{
  PyObject *k = 0, *v = 0, *rc = 0;

  if ((k = PyIter_Next(ITER_I(me))) != 0 &&
      (v = PyObject_GetItem(ITER_MAP(me), k)) != 0)
    rc = Py_BuildValue("(OO)", k, v);
  Py_XDECREF(k); Py_XDECREF(v);
  return (rc);
}

static const PyTypeObject itemiter_pytype_skel = {
  PyObject_HEAD_INIT(0) 0,		/* Header */
  "ItemIter",				/* @tp_name@ */
  sizeof(iter_pyobj),			/* @tp_basicsize@ */
  0,					/* @tp_itemsize@ */

  iter_pydealloc,			/* @tp_dealloc@ */
  0,					/* @tp_print@ */
  0,					/* @tp_getattr@ */
  0,					/* @tp_setattr@ */
  0,					/* @tp_compare@ */
  0,					/* @tp_repr@ */
  0,					/* @tp_as_number@ */
  0,					/* @tp_as_sequence@ */
  0,					/* @tp_as_mapping@ */
  0,					/* @tp_hash@ */
  0,					/* @tp_call@ */
  0,					/* @tp_str@ */
  0,					/* @tp_getattro@ */
  0,					/* @tp_setattro@ */
  0,					/* @tp_as_buffer@ */
  Py_TPFLAGS_DEFAULT |			/* @tp_flags@ */
    Py_TPFLAGS_BASETYPE,

  /* @tp_doc@ */
  "Iterates over the items of a mapping.",

  0,					/* @tp_traverse@ */
  0,					/* @tp_clear@ */
  0,					/* @tp_richcompare@ */
  0,					/* @tp_weaklistoffset@ */
  PyObject_SelfIter,			/* @tp_iter@ */
  itemiter_pynext,			/* @tp_iternext@ */
  0,					/* @tp_methods@ */
  0,					/* @tp_members@ */
  0,					/* @tp_getset@ */
  0,					/* @tp_base@ */
  0,					/* @tp_dict@ */
  0,					/* @tp_descr_get@ */
  0,					/* @tp_descr_set@ */
  0,					/* @tp_dictoffset@ */
  0,					/* @tp_init@ */
  PyType_GenericAlloc,			/* @tp_alloc@ */
  abstract_pynew,			/* @tp_new@ */
  0,					/* @tp_free@ */
  0					/* @tp_is_gc@ */
};

static PyObject *valiter_pynext(PyObject *me)
{
  PyObject *k = 0, *rc = 0;

  if ((k = PyIter_Next(ITER_I(me))) != 0)
    rc = PyObject_GetItem(ITER_MAP(me), k);
  Py_XDECREF(k);
  return (rc);
}

static const PyTypeObject valiter_pytype_skel = {
  PyObject_HEAD_INIT(0) 0,		/* Header */
  "ValueIter",				/* @tp_name@ */
  sizeof(iter_pyobj),			/* @tp_basicsize@ */
  0,					/* @tp_itemsize@ */

  iter_pydealloc,			/* @tp_dealloc@ */
  0,					/* @tp_print@ */
  0,					/* @tp_getattr@ */
  0,					/* @tp_setattr@ */
  0,					/* @tp_compare@ */
  0,					/* @tp_repr@ */
  0,					/* @tp_as_number@ */
  0,					/* @tp_as_sequence@ */
  0,					/* @tp_as_mapping@ */
  0,					/* @tp_hash@ */
  0,					/* @tp_call@ */
  0,					/* @tp_str@ */
  0,					/* @tp_getattro@ */
  0,					/* @tp_setattro@ */
  0,					/* @tp_as_buffer@ */
  Py_TPFLAGS_DEFAULT |			/* @tp_flags@ */
    Py_TPFLAGS_BASETYPE,

  /* @tp_doc@ */
  "Iterates over the values of a mapping.",

  0,					/* @tp_traverse@ */
  0,					/* @tp_clear@ */
  0,					/* @tp_richcompare@ */
  0,					/* @tp_weaklistoffset@ */
  PyObject_SelfIter,			/* @tp_iter@ */
  valiter_pynext,			/* @tp_iternext@ */
  0,					/* @tp_methods@ */
  0,					/* @tp_members@ */
  0,					/* @tp_getset@ */
  0,					/* @tp_base@ */
  0,					/* @tp_dict@ */
  0,					/* @tp_descr_get@ */
  0,					/* @tp_descr_set@ */
  0,					/* @tp_dictoffset@ */
  0,					/* @tp_init@ */
  PyType_GenericAlloc,			/* @tp_alloc@ */
  abstract_pynew,			/* @tp_new@ */
  0,					/* @tp_free@ */
  0					/* @tp_is_gc@ */
};

const PySequenceMethods gmap_pysequence = {
  0,					/* @sq_length@ */
  0,					/* @sq_concat@ */
  0,					/* @sq_repeat@ */
  0,					/* @sq_item@ */
  0,					/* @sq_slice@ */
  0,					/* @sq_ass_item@ */
  0,					/* @sq_ass_slice@ */
  PyMapping_HasKey,			/* @sq_contains@ */
  0,					/* @sq_inplace_concat@ */
  0					/* @sq_inplace_repeat@ */
};

/*----- Other mapping protocol support ------------------------------------*/

Py_ssize_t gmap_pysize(PyObject *me)
{
  PyObject *i = 0, *x = 0;
  Py_ssize_t rc = -1, n = 0;

  if ((i = PyObject_GetIter(me)) == 0) goto done;
  while ((x = PyIter_Next(i)) != 0) { n++; Py_DECREF(x); x = 0; }
  if (PyErr_Occurred()) goto done;
  rc = n;
done:
  Py_XDECREF(i); Py_XDECREF(x);
  return (rc);
}

PyObject *gmapmeth_has_key(PyObject *me, PyObject *arg)
{
  PyObject *k;
  if (!PyArg_ParseTuple(arg, "O:has_key", &k)) return (0);
  return (getbool(PyMapping_HasKey(me, k)));
}

PyObject *gmapmeth_keys(PyObject *me)
{
  PyObject *l = 0, *i = 0, *k, *rc = 0;
  int err;

  if ((l = PyList_New(0)) == 0 ||
      (i = PyObject_GetIter(me)) == 0)
    goto done;
  while ((k = PyIter_Next(i)) != 0)
    { err = PyList_Append(l, k); Py_DECREF(k); if (err) goto done; }
  if (PyErr_Occurred()) goto done;
  rc = l; l = 0;
done:
  Py_XDECREF(l); Py_XDECREF(i);
  return (rc);
}

PyObject *gmapmeth_values(PyObject *me)
{
  PyObject *l = 0, *i = 0, *k, *v, *rc = 0;
  int err = 0;

  if ((l = PyList_New(0)) == 0 ||
      (i = PyObject_GetIter(me)) == 0)
    goto done;
  while ((k = PyIter_Next(i)) != 0) {
    if ((v = PyObject_GetItem(me, k)) == 0 ||
	PyList_Append(l, v))
      err = -1;
    Py_DECREF(k); Py_XDECREF(v);
    if (err) goto done;
  }
  if (PyErr_Occurred()) goto done;
  rc = l; l = 0;
done:
  Py_XDECREF(l); Py_XDECREF(i);
  return (rc);
}

PyObject *gmapmeth_items(PyObject *me)
{
  PyObject *l = 0, *i = 0, *k, *v, *z, *rc = 0;
  int err = 0;

  if ((l = PyList_New(0)) == 0 ||
      (i = PyObject_GetIter(me)) == 0)
    goto done;
  while ((k = PyIter_Next(i)) != 0) {
    z = 0;
    if ((v = PyObject_GetItem(me, k)) == 0 ||
	(z = Py_BuildValue("(OO)", k, v)) == 0 ||
	PyList_Append(l, z))
      err = -1;
    Py_DECREF(k); Py_XDECREF(v); Py_XDECREF(z);
    if (err) goto done;
  }
  if (PyErr_Occurred()) goto done;
  rc = l; l = 0;
done:
  Py_XDECREF(l); Py_XDECREF(i);
  return (rc);
}

PyObject *gmapmeth_iterkeys(PyObject *me)
  { return (PyObject_GetIter(me)); }

PyObject *gmapmeth_itervalues(PyObject *me)
{
  PyObject *i;
  iter_pyobj *ii;

  if ((i = PyObject_GetIter(me)) == 0)
    return (0);
  ii = PyObject_NEW(iter_pyobj, valiter_pytype);
  ii->map = me; Py_INCREF(me);
  ii->i = i;
  return ((PyObject *)ii);
}

PyObject *gmapmeth_iteritems(PyObject *me)
{
  PyObject *i;
  iter_pyobj *ii;

  if ((i = PyObject_GetIter(me)) == 0)
    return (0);
  ii = PyObject_NEW(iter_pyobj, itemiter_pytype);
  ii->map = me; Py_INCREF(me);
  ii->i = i;
  return ((PyObject *)ii);
}

PyObject *gmapmeth_clear(PyObject *me)
{
  PyObject *i = 0, *k = 0, *rc = 0;

  if ((i = PyObject_GetIter(me)) == 0)
    goto end;
  while ((k = PyIter_Next(i)) != 0) {
    PyObject_DelItem(me, k);
    Py_DECREF(k);
  }
  if (PyErr_Occurred()) goto end;
  rc = me; Py_INCREF(me);
end:
  Py_XDECREF(i);
  return (rc);
}

static const char *const def_kwlist[] = { "key", "default", 0 };

PyObject *gmapmeth_get(PyObject *me, PyObject *arg, PyObject *kw)
{
  PyObject *k, *def = Py_None, *v;

  if (!PyArg_ParseTupleAndKeywords(arg, kw, "O|O:get",
				   (/*unconst*/ char **)def_kwlist,
				   &k, &def))
    return (0);
  if ((v = PyObject_GetItem(me, k)) != 0) return (v);
  PyErr_Clear();
  RETURN_OBJ(def);
}

PyObject *gmapmeth_setdefault(PyObject *me, PyObject *arg, PyObject *kw)
{
  PyObject *k, *def = Py_None, *v;

  if (!PyArg_ParseTupleAndKeywords(arg, kw, "O|O:setdefault",
				   (/*unconst*/ char **)def_kwlist,
				   &k, &def))
    return (0);
  if ((v = PyObject_GetItem(me, k)) != 0) return (v);
  PyErr_Clear();
  if (PyObject_SetItem(me, k, def)) return (0);
  RETURN_OBJ(def);
}

PyObject *gmapmeth_pop(PyObject *me, PyObject *arg, PyObject *kw)
{
  PyObject *k, *def = 0, *v;

  if (!PyArg_ParseTupleAndKeywords(arg, kw, "O|O:pop",
				   (/*unconst*/ char **)def_kwlist,
				   &k, &def))
    return (0);
  if ((v = PyObject_GetItem(me, k)) != 0) {
    PyObject_DelItem(me, k);
    return (v);
  } else if (def) {
    PyErr_Clear();
    RETURN_OBJ(def);
  } else
    return (0);
}

PyObject *gmapmeth_update(PyObject *me, PyObject *arg)
{
  PyObject *map, *i = 0, *k, *v, *rc = 0;
  int err = 0;

  if (!PyArg_ParseTuple(arg, "O:update", &map) ||
      (i = PyObject_GetIter(map)) == 0)
    goto end;
  while ((k = PyIter_Next(i)) != 0) {
    if ((v = PyObject_GetItem(map, k)) == 0 ||
	PyObject_SetItem(me, k, v))
      err = -1;
    Py_DECREF(k); Py_XDECREF(v);
    if (err) goto end;
  }
  if (PyErr_Occurred()) goto end;
  rc = me; Py_INCREF(me);
end:
  Py_XDECREF(i);
  return (rc);
}

PyObject *gmapmeth_popitem(PyObject *me)
{
  PyObject *i = 0, *k = 0, *v = 0, *rc = 0;

  if ((i = PyObject_GetIter(me)) == 0)
    goto end;
  if ((k = PyIter_Next(i)) == 0) {
    if (!PyErr_Occurred()) VALERR("popitem(): mapping is empty");
    goto end;
  }
  if ((v = PyObject_GetItem(me, k)) == 0 ||
      PyObject_DelItem(me, k))
    goto end;
  rc = Py_BuildValue("(OO)", k, v);
end:
  Py_XDECREF(i); Py_XDECREF(k); Py_XDECREF(v);
  return (rc);
}

const PyMethodDef gmap_pymethods[] = {
  GMAP_METHODS
  { 0 }
};

/*----- Submodule initialization ------------------------------------------*/

void pyke_gmap_pyinit(void)
{
  INITTYPE(itemiter, root);
  INITTYPE(valiter, root);
}

void pyke_gmap_pyinsert(PyObject *mod)
{
  INSERT("ItemIter", itemiter_pytype);
  INSERT("ValueIter", valiter_pytype);
}

/*----- That's all, folks -------------------------------------------------*/
